import os
import subprocess
import pkg_resources
from .nbsearch import _NBSEARCH_DB_PATH, create_init_db
import subprocess
from datasette import hookimpl

__version__="0.0.4"


@hookimpl
def extra_css_urls(database, table, columns, view_name, datasette):
    return [
        #"https://cdnjs.cloudflare.com/ajax/libs/prism/1.22.0/themes/prism.min.css",
        #"https://cdnjs.cloudflare.com/ajax/libs/prism/1.22.0/plugins/line-numbers/prism-line-numbers.min.css",
        "/-/static-plugins/nbsearch/prism.css",
        "/-/static-plugins/nbsearch/nbsearch.css",
    ]


@hookimpl
def extra_js_urls(database, table, columns, view_name, datasette):
    return [
        #"https://cdn.jsdelivr.net/npm/prismjs@1.22.0/prism.min.js",
        #"https://cdnjs.cloudflare.com/ajax/libs/prism/1.22.0/plugins/line-numbers/prism-line-numbers.min.js",
        #"https://cdnjs.cloudflare.com/ajax/libs/prism/1.22.0/components/prism-python.min.js",
        #"https://cdn.jsdelivr.net/npm/marked/marked.min.js",
        #"https://cdn.jsdelivr.net/npm/clipboard@2.0.6/dist/clipboard.min.js",
        "/-/static-plugins/nbsearch/prism.js",
        "/-/static-plugins/nbsearch/marked.min.js",
        "/-/static-plugins/nbsearch/clipboard.min.js",
    ]

def setup_nbsearch():
    # Spawn a process somewhere to initialise the indexing
    #subprocess.run(["nbsearch", "index"])
    # Make sure tables are created
    create_init_db()
    subprocess.run(["nbsearch", "index"])

    #fpath = pkg_resources.resource_filename('nbsearch', '/static/')
    return {
        "command": [
            "datasette",
            "serve",
            f"{_NBSEARCH_DB_PATH}",
            #f"--template-dir={fpath}templates/",
            #"--metadata",
            #f"{fpath}metadata/metadata.json",
            #"--static",
            #f"static:{fpath}static/",
            "-p",
            "{port}",
            "--config",
            "base_url:{base_url}nbsearch/"
        ],
        "absolute_url": True,
        # The following needs a the labextension installing.
        # eg in postBuild: jupyter labextension install jupyterlab-server-proxy
        "launcher_entry": {
            "enabled": True,
            #'icon_path': f'{fpath}/icons/nbsearch.svg',
            "title": "nbsearch",
        },
    }



# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.6.0
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # `nbsearch`
#
# Simple sqlite/datasette powered full text search engine for notebooks.
# Based on: Simon Willison's ["Fast Autocomplete Search for Your Website"](https://24ways.org/2018/fast-autocomplete-search-for-your-website/).
#
# This file provides tools for creating and updating the database.

# %%
# #%pip install --upgrade sqlite_utils
# #%pip install --upgrade git+https://github.com/innovationOUtside/nb_quality_profile.git           
import nbformat
import sqlite3
import jupytext
import sqlite_utils
import os
import hashlib
import uuid
#from tqdm import tqdm
from nb_quality_profile import nb_visualiser as nbv
from pathlib import Path
from sqlite_utils import Database

_NBSEARCH_USER_PATH = os.path.join(str(Path.home()), ".nbsearch") 
# The indexing should really take place over directories the notebook server sees, but:
# - we can't reliably see JUPYTER_ROOT_DIRECTORY ?
# - we can't access c.NotebookApp.notebook_dir ?
_NB_SEARCH_PATH = os.path.join(str(Path.home()))
_NBSEARCH_DB_FILE = "notebooks.sqlite"
_NBSEARCH_DB_PATH =  os.path.join(_NBSEARCH_USER_PATH, _NBSEARCH_DB_FILE)
_FILES_TABLE = "nbfiles"
_CONTENTS_TABLE = "nbcontents"

if not os.path.exists(_NBSEARCH_USER_PATH):
    os.makedirs(_NBSEARCH_USER_PATH)

def nbpathwalk(path, filetypes=None):
    ''' Walk down a directory path looking for ipynb notebook files… '''
    for path, _, files in os.walk(path):
        # Omit hidden directories
        if '/.' in path:
            continue
        # TO DO — allow other file types
        # that can be parsed with Jupytext
        if filetypes is None:
            filetypes = ('.ipynb')
        for f in [i for i in files if i.endswith(tuple(filetypes))]:
            yield os.path.join(path, f)

def get_nb(fn, text_formats=False):
    """Get notebook contents."""
    
    fmts = ['.ipynb']
    if text_formats:
        fmts = fmts + ['.md', '.Rmd', '.py']
    _fn, fn_ext = os.path.splitext(fn)

    if fn_ext not in fmts or not os.path.isfile(fn):
        # Better to return this as empty and check downstream?
        return {}

    if fn_ext=='.ipynb':
        with open(fn, 'r') as f:
            try:
                nb = nbformat.reads(f.read(), nbformat.NO_CONVERT)
            except:
                print(f"Couldn't process {fn}")
                return {}
    else:
        try:
            nb = jupytext.read(fn)
        except:
            print(f"Couldn't process {fn}")
            return {}
    return nb
    
def get_cell_contents(nb, cell_typ=None):
    ''' Extract the content of Jupyter notebook cells. '''
    
    if not nb or "cells" not in nb:
        return []

    if cell_typ is None:
        cell_typ = ['markdown', 'code', 'raw']
    
    cells = [i for i in nb.cells if i['cell_type'] in cell_typ]

    return cells


def index_notebook(nbid, nb, cell_typ=None, text_formats=False, no_img = True):
    """ Parse individual notebook."""
    
    docs = []
    cells = get_cell_contents(nb, cell_typ=cell_typ)
    cnt = {'all':0, 'markdown':0, 'code':0, 'raw':0}

    # TO DO - the plotter should have: plt.ioff() ?
    # Currently raising: ApplePersistenceIgnoreState: Existing state will not be touched.
    # This is brittle - matplotlib insists on trying to open a gui
    if not no_img:
        img = nbv.nb_vis_parse_nb(raw=nb, minimal=True, retval='img') if nb else None
    else:
        img = None

    for cell in cells:
        
        if cell['cell_type'] in cnt:
            cnt[cell['cell_type']] += 1
            cell['cell_type_num'] = cnt[cell['cell_type']]
        else:
            cell['cell_type_num'] = 0
        cell['tags'] = cell['metadata']['tags'] if 'tags' in cell['metadata'] else ''
        cell['nbid'] = nbid
        cell['cell_num'] = cnt['all']
        
        cnt['all'] += 1
        docs.append({k: cell[k] for k in ('nbid', 'source', 'cell_type',
                                          'tags', 'cell_num', 'cell_type_num')})
    
    return docs, cnt, img


def create_tables(db, files_table=_FILES_TABLE, contents_table=_CONTENTS_TABLE):
    """Create database tables."""
    if files_table not in db.table_names():
        db[files_table].create({"nbid": str, "last_modified": int, 'cells': int,
                                "md_cells": int, "code_cells": int,
                                "name": str, "file_type": str,
                                "img": "BLOB"}, pk="nbid")

    if contents_table not in db.table_names():
        db[contents_table].create({"nbid": str, "tags": str,
                                   "source": str, "cell_type": str, "cell_num": int,
                                   "cell_type_num": int},
                                   pk=("nbid", "cell_num"),
                                   foreign_keys=[ ("nbid", files_table, "nbid")])


def create_init_db(dbpath=_NBSEARCH_DB_PATH, clear=False):
    """Create and initialise database."""
    if clear and os.path.exists(dbpath):
        os.remove(dbpath)
    db = sqlite_utils.Database(dbpath)
    create_tables(db)


def remove_notebook(db, nbid=None, fn=None,
                    files_table=_FILES_TABLE, contents_table=_CONTENTS_TABLE):
    """Remove a notebook from the database."""
    # TO DO - do we also have to monitor this from file system? eg file deletions?
    # https://github.com/gorakhargosh/watchdog/ could be handy?
    # And s/thing like: nohup python /path/to/watchdog.py &
    if isinstance(db, str) and os.path.isfile(db):
        db = Database(db)
    nbid = get_nbid(nbid, fn, uid=False)
    db[contents_table].delete_where(f"nbid = '{nbid}'")
    db[files_table].delete_where(f"nbid = '{nbid}'")
    
def get_nbid(nbid=None, fn=None, uid=True):
    """Create notebook id."""
    # Create a uid, ideally based on the notebook name if not already provided
    if ((not nbid) and fn):
        nbid = hashlib.md5(fn.encode()).hexdigest()
    elif uid:
        nbid = str(uuid.uuid4())
    return nbid


def update_notebook(db, nbid=None, fn=None, nbcontent=None, 
                    files_table=_FILES_TABLE, contents_table=_CONTENTS_TABLE,
                    cell_typ=None, text_formats=None, fts_update=False):
    """Update items relating to a notebook."""
    if isinstance(db, str) and os.path.isfile(db):
        db = Database(db)
    
    if not nbid and not fn and not nbcontent:
        return

    nbid = get_nbid(nbid, fn)
    #print(f"remove...{nbid}")
    remove_notebook(db, nbid, files_table, contents_table)

    nb = nbcontent if nbcontent else get_nb(fn, text_formats=text_formats)
    if not nb:
        return

    _fn, fn_ext = os.path.splitext(fn)
    docs, cnt, img = index_notebook(nbid, nb, cell_typ=cell_typ)
    f_details = {"nbid": nbid, "last_modified": os.path.getmtime(fn),
                 "cells": cnt['all'],
                 "md_cells": cnt['markdown'], "code_cells": cnt['code'],
                 "name": fn, "file_type": fn_ext,
                 "img": img}
    db[files_table].upsert(f_details, pk="nbid")
    db[contents_table].upsert_all(docs, pk=("nbid", "cell_num"))

    if fts_update:
        update_fts(db, contents_table)

def update_fts(db, contents_table=_CONTENTS_TABLE):
    """Update full text search."""
    if not db[contents_table].detect_fts():
        db[contents_table].enable_fts(["source"])
    else:
        db[contents_table].populate_fts(["source"])

# TO DO - at the moment, if we delete a notebook
# it still exists in the search database
def index_notebooks_sqlite(nbpath='.', dbpath=_NBSEARCH_DB_PATH, cell_typ=None,
                           files_table=_FILES_TABLE,
                           contents_table=_CONTENTS_TABLE,
                           text_formats=None):
    ''' Get content from each notebook down a path and index it. '''
    
    db = sqlite_utils.Database(dbpath)
    create_tables(db, files_table, contents_table)
    for fn in nbpathwalk(nbpath):
        update_notebook(db, fn=fn, cell_typ=cell_typ, text_formats=text_formats)
    
    update_fts(db, contents_table)
        
    # TO DO
    # Notebook should update record in db whe notebook saves
    # nbautoexport has a way of updating c.FileContentsManager.post_save_hook
    # in a way that does not break previously added hooks
    # (Is there no native safe Jupyter way of doing this? What does Jupytext do?)
    # https://github.com/drivendataorg/nbautoexport/blob/master/nbautoexport/jupyter_config.py

# %% tags=["active-ipynb"]
# #!rm notebooks.sqlite
# create_init_db(clear=True)

# %% tags=["active-ipynb"]
# path = '/Users/tonyhirst/Documents/GitHub/tm129-robotics2020/content/01. Introducing notebooks and the RoboLab environment'
# os.listdir(path)

# %% tags=["active-ipynb"]
# #index_notebooks_sqlite('01. Introducing notebooks and the RoboLab environment')
# index_notebooks_sqlite(path)

# %% tags=["active-ipynb"]
# db = sqlite_utils.Database('notebooks.sqlite')
# #list(db['nbcontents_fts'].rows)

# %%
# TO DO - make jupyter-server-proxy
# TO DO - make update index button
# TO DO - make updates only index changed things
# TO DO - delete items not in index and more


# %% tags=["active-ipynb"]
# #jupyter_notebook_config.py
# # Derived from: https://jupyter-notebook.readthedocs.io/en/stable/extending/savehooks.html
# import os
# from nbsearch.nbsearch import update_notebook
#
# _script_exporter = None
#
# def db_post_save(model, os_path, contents_manager, **kwargs):
#     """Update nbsearch db."""
#     from nbconvert.exporters.script import ScriptExporter
#
#     if model['type'] != 'notebook':
#         return
#
#     # Useful? contents_manager.root_dir
#
#     update_notebook(db, nbid=None, fn=os_path, nbcontent=model['content'])
#
# c.FileContentsManager.post_save_hook = db_post_save



import click
import subprocess
import pkg_resources
from .nbsearch import index_notebooks_sqlite
from .nbsearch import _NBSEARCH_DB_PATH, _NB_SEARCH_PATH
from .nbwatchdog import dbmonitor
import os
from nbsearch.nbsearch import create_init_db
from nbsearch.nbsearch import _NBSEARCH_USER_PATH, _NBSEARCH_DB_PATH

@click.group()
def cli():
	pass

@cli.command()
@click.option('--path', '-p', default=_NBSEARCH_DB_PATH)
@click.option('--clear/--no-clear', default=False)
def create(path, clear):
    """Create inital empty db."""
    create_init_db(path, clear)


@cli.command()
@click.option('--path', '-p', default='.', type=click.Path(exists=True))
def index(path):
    """Index path."""
    click.echo('Indexing file/directory: {}'.format(path))
    # Create some useful paths
    if not os.path.exists(_NBSEARCH_USER_PATH):
        os.makedirs(_NBSEARCH_USER_PATH)
    # Would be useful to export these as persistent environment variables?
    create_init_db(_NBSEARCH_DB_PATH)
    index_notebooks_sqlite(path)

@cli.command()
@click.option('--dbpath', '-p', default=_NBSEARCH_DB_PATH, type=click.Path(exists=True))
def serve(dbpath):
    """Run server path."""
    #fpath = pkg_resources.resource_filename('nbsearch', '/static/') #static/
    command =  [
            "datasette",
            "serve",
            dbpath,
            #f"--template-dir={fpath}templates/",
            #"--metadata",
            #f"{fpath}metadata/metadata.json",
            #"--static",
            #f"static:{fpath}static/",
            "-p",
            "0", # pick a port for me...
            #"--config",
            #"base_url:{base_url}nbsearch/",
            #"-d",
            #os.environ["HOME"]
        ]
   
    subprocess.Popen(command)

# Not tested
@cli.command()
@click.option('--searchpath', '-s', default=_NB_SEARCH_PATH, type=click.Path(exists=True))
@click.option('--path', '-p', default=_NBSEARCH_DB_PATH)
@click.option('--clear/--no-clear', default=False)
def monitor(searchpath, path, clear):
    """Monitor notebook path."""
    create_init_db(path, clear)
    dbmonitor(searchpath)
    # nohup nbsearch monitor &






#http://thepythoncorner.com/dev/how-to-create-a-watchdog-in-python-to-look-for-filesystem-changes/

from .nbsearch import update_notebook, remove_notebook, _NBSEARCH_DB_PATH
import time
from watchdog.observers import Observer
from watchdog.events import RegexMatchingEventHandler

db = _NBSEARCH_DB_PATH

def on_created(event):
    #print(f'created {event.src_path}')
    update_notebook(db, fn=event.src_path)

def on_deleted(event):
    #print(f'deleted {event.src_path}')
    remove_notebook(db, fn=event.src_path)

def on_modified(event):
    #print(f'modified {event.src_path}')
    update_notebook(db, fn=event.src_path, fts_update=True)

def on_moved(event):
    #print(f'moved {event.src_path}, {event.dest_path}')
    remove_notebook(db, fn=event.src_path)
    update_notebook(db, fn=event.dest_path, fts_update=True)

regexes = [r".*\.py", r".*\.ipynb", r".*\.Rmd", r".*\.md"]
ignore_regexes = [r".*[/\\]\..*"]

ignore_directories = False # Not sure I understand the semantics of this?
case_sensitive = True
event_handler = RegexMatchingEventHandler(regexes, ignore_regexes, ignore_directories, case_sensitive)
event_handler.on_created = on_created
event_handler.on_deleted = on_deleted
event_handler.on_modified = on_modified
event_handler.on_moved = on_moved

def dbmonitor(path='.'):
    """Monitor a path."""
    #print(f"Starting to monitor {path}")
    observer = Observer()
    observer.schedule(event_handler, path, recursive=True)
    while True:
        observer.start()
        try:
            while True:
                time.sleep(1)
        except:
            observer.stop()
    observer.join()


from setuptools import setup
import os
from os import path
from pathlib import Path

def get_long_description():
    with open(
        path.join(path.dirname(path.abspath(__file__)), "README.md"),
        encoding="utf8",
    ) as fp:
        return fp.read()

def get_requirements(fn='requirements.txt', nogit=True):
   """Get requirements."""
   if path.exists(fn):
      with open(fn, 'r') as f:
        requirements = f.read().splitlines()
   else:
     requirements = []
   requirements = [r.split()[0].strip() for r in requirements if r and not r.startswith('#')]
   if nogit:
       requirements = [r for r in requirements if not r.startswith('git+')]
   return requirements
   
requirements = get_requirements()

print(f'Requirements: {requirements}')

extras = {
    }



setup(
    # Meta
    author='Tony Hirst',
    author_email='tony.hirst@open.ac.uk',
    name='nbsearch',
    description='datasette powered notebook search',
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    license='MIT License',
    url="https://github.com/ouseful-testing/nbsearch",
    # Dependencies
    install_requires=requirements,
    #setup_requires=[],
    extras_require=extras,

    # Packaging
    entry_points={
        'jupyter_serverproxy_servers': [
        'nbsearch = nbsearch:setup_nbsearch',
      ],
      "datasette": ["nbsearch = nbsearch"],
      'console_scripts': ['nbsearch = nbsearch.cli:cli']},
    include_package_data=True,
    package_data={
        "nbsearch": ["static/prism.js",
         "static/clipboard.min.js",
         "static/prism.css",
         "static/marked.min.js",
          "static/nbsearch.css",
         "templates/index.html",
         "static/thebelab.js"
         ],
    },
    zip_safe=False,
    packages=['nbsearch'],

    # Classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Education',
        'License :: Free For Educational Use',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Education'
    ],
)


"""
from nbsearch.nbsearch import create_init_db
from nbsearch.nbsearch import _NBSEARCH_USER_PATH, _NBSEARCH_DB_PATH
# Create some useful paths
if not os.path.exists(_NBSEARCH_USER_PATH):
    os.makedirs(_NBSEARCH_USER_PATH)
# Would be useful to export these as persistent environment variables?
create_init_db(_NBSEARCH_DB_PATH)
"""